/******************************************************************************
  Name: compile-l.pl
  Description: This file consists of two main predicates for converting
    Prholog code into Prolog code. The first predicate 
    compile_rholog_clause(+Prholog_clause) converts code given by .rho files.
    The second one ?(+Query,-Result) is the top-level predicate that executes 
    Rholog goals. Additionally it contains predicates for handling regular
    expressions.
  Version: 1.0.0.0
  Author(s): Besik Dundua
  Created: 2009-09-09
  Last edited: 2015-11-17
  Last edited by: Klaus Reisenberger-Hagmayer
  Copyright:
  License:
  Notes:
    Note 1: here we have to define all strategies of the library.
    Note 2: here we store all strategy names which use the operator  ==>
    Note 3: here we store all strategy names which use the neg operator  =\=>
    Note 4: 
    Note 5: if the variable is not found (Out = false) the program gets stuck 
    Note 6: either this call fails or returns a value for Out
    Note 7: if Out equals the empty set (or hole for contexts) we succeeded 
      otherwise the regex matches the input only up to a certain point
    Note8: is needed because seq_var crashes if Pattern is not an atom
******************************************************************************/
%------------------------------------------------------------------------------
%% compile_rholog_clause(+Prholog_clause)
%
% Converts the clause given by Prholog_clause to Prolog code
%------------------------------------------------------------------------------
compile_rholog_clause(end_of_file):-
    !.
compile_rholog_clause(Rholog_clause):-
    break_point,
    compile_rholog_clause_to_prolog(Rholog_clause, Prolog_clause),
    !,
    portray_clause(Prolog_clause).

%------------------------------------------------------------------------------
% compile_rholog_clause_to_prolog(+Prholog_clause, -Prolog_clause)
% Converts the given Prholog clause to a Prolog clause. Which case is used 
% depends whether or not it is a negative strategy, there is a body or there
% is a regular constraint 
%------------------------------------------------------------------------------
compile_rholog_clause_to_prolog(
    (Strategy :: Lhs ==> Rhs :- Body), Prolog_clause
) :-
    !,
    compile_rholog_clause_to_prolog(
        (Strategy :: Lhs ==> Rhs where _:- Body), Prolog_clause).
compile_rholog_clause_to_prolog(
    (Strategy :: Lhs ==> Rhs where Constraints:- Body), 
    (Clause_head :- Prolog_body_out)
) :-
    !,
    hedge_to_term_lists_and_appends(
        Lhs, [], true, [], Subst1, Appends1, Lhs_list),
    hedge_to_term_lists_and_appends(
        Rhs, Subst1, true, [],  Subst2, Appends2, Rhs_list),
    get_strategy_name_and_arguments(
        Strategy, Subst2, Str_name, Str_args_list, Subst3, Appends1, Appends3),
    add_strategy(Str_name),
    assert(to_abolish_rholog_internal(Str_name,6)),
    compile_rholog_clause_body(Body, Subst3, _, Prolog_body_in),
    Clause_head =.. [
        Str_name, Str_args_list, Lhs_list, Rhs_list,Constraints
    ],
    remove_true(
        (Appends3,check_constraints(Subst3,Constraints)), Prolog_body_in, 
        (Appends2,check_constraints(Subst3,Constraints)), Prolog_body_out).
compile_rholog_clause_to_prolog((Strategy :: Lhs ==> Rhs), (Prolog_clause)) :-
    !,
    compile_rholog_clause_to_prolog(
        (Strategy :: Lhs ==> Rhs where _), (Prolog_clause)).
compile_rholog_clause_to_prolog(
    (Strategy :: Lhs ==> Rhs where Constraints), (Prolog_clause)
) :-
    !,
    hedge_to_term_lists_and_appends(
        Lhs, [], true, [], Subst1, Appends1, Lhs_list),
    hedge_to_term_lists_and_appends(
        Rhs, Subst1, true, [],  Subst2, Appends2, Rhs_list),
    get_strategy_name_and_arguments(
        Strategy, Subst2, Str_name, Str_args_list, Subst3, Appends1, Appends3),
    add_strategy(Str_name),
    assert(to_abolish_rholog_internal(Str_name,6)),
    Clause_head =.. [Str_name, Str_args_list, Lhs_list, Rhs_list, Constraints],
    make_prolog_clause(
        Clause_head, (Appends3,check_constraints(Subst3,Constraints)),
        (Appends2,check_constraints(Subst3,Constraints)), Prolog_clause).
compile_rholog_clause_to_prolog(
    (Strategy_shortcut := Strategy_full), Compiled
) :-
    !,
    compile_rholog_clause_to_prolog(
        (Strategy_shortcut :: s_X ==> s_Y :- Strategy_full :: s_X ==> s_Y),
        Compiled).
compile_rholog_clause_to_prolog((:- X), (:- X)):-
    X,
    !.
compile_rholog_clause_to_prolog(Prolog_clause, Prolog_clause).

%------------------------------------------------------------------------------
% compile_rholog_clause_body(
%     +Prholog_body, +Substitution_in, Substitution_out, -Prolog_body)
% Converts the given Prholog body to a Prolog body extending Substitution_in
% with variables (individual, function, sequent) which occur in Prholog
%------------------------------------------------------------------------------

compile_rholog_clause_body(
    Strategy :: Lhs ==> Rhs, Subst_in, Subst_out, Trans
) :-
    compile_rholog_clause_body(
        Strategy :: Lhs ==> Rhs where _, Subst_in, Subst_out, Trans).
compile_rholog_clause_body(
    Strategy :: Lhs ==> Rhs where Constraints, Subst_in, Subst_out, Trans
) :-
    get_strategy_name_and_arguments(
        Strategy, Subst_in, Str_name, Str_args, Subst_temp1, [], Appends3),
    % add_strategy(Str_name), %TBD: Check whether this line can be removed
    !,
    hedge_to_term_lists_and_appends(
        Lhs, Subst_temp1, true, Appends3, Subst_temp2, Appends1, Lhs_list),
    hedge_to_term_lists_and_appends(
        Rhs, Subst_temp2, true, [], Subst_out, Appends2, Rhs_list),
    get_translated_literal(
        positive, Str_name, Str_args, Lhs_list, Rhs_list, Constraints,
        Trans_Temp),
    remove_true(
        (Appends1,check_constraints(Subst_out,Constraints)), Trans_Temp,
        (Appends2,check_constraints(Subst_out,Constraints)), Trans).
compile_rholog_clause_body(
    Strategy :: Lhs =\=> Rhs, Subst_in, Subst_out, Trans
) :-
    compile_rholog_clause_body(
        Strategy :: Lhs =\=> Rhs where _, Subst_in, Subst_out, Trans).
compile_rholog_clause_body(
    Strategy :: Lhs =\=> Rhs where Constraints, Subst_in, Subst_out, Trans
) :-
    get_strategy_name_and_arguments(Strategy, Subst_in, Str_name, 
        Str_args, Subst_temp1, [], Appends3),
    % add_neg_strategy(Str_name),
    !,
    hedge_to_term_lists_and_appends(Lhs, Subst_temp1, true, Appends3, 
        Subst_temp2, Appends1, Lhs_list),
    hedge_to_term_lists_and_appends(Rhs, Subst_temp2, true, [],  Subst_out,
        Appends2, Rhs_list),
    get_translated_literal(negative, Str_name, Str_args, Lhs_list, Rhs_list,
        Constraints, Trans_Temp),
    remove_true(
        (Appends1,check_constraints(Subst_out,Constraints)), Trans_Temp,
        (Appends2,check_constraints(Subst_out,Constraints)), Trans).
compile_rholog_clause_body(
    (Rholog_subgoal1, Rholog_subgoal2), Subst_in, Subst_out,
    (Prolog_subgoal1, Prolog_subgoal2)
) :-
    !,
    compile_rholog_clause_body(
        Rholog_subgoal1,  Subst_in, Subst, Prolog_subgoal1),
    compile_rholog_clause_body(
        Rholog_subgoal2,  Subst, Subst_out, Prolog_subgoal2).
compile_rholog_clause_body(
    (Rholog_subgoal1 -> Rholog_subgoal2), Subst_in, Subst_out,
    (Prolog_subgoal1 -> Prolog_subgoal2)
) :-
    !,
    compile_rholog_clause_body(
        Rholog_subgoal1, Subst_in, Subst, Prolog_subgoal1),
    compile_rholog_clause_body(
        Rholog_subgoal2,  Subst, Subst_out, Prolog_subgoal2).
compile_rholog_clause_body(
    (Rholog_subgoal1 *-> Rholog_subgoal2), Subst_in, Subst_out,
    (Prolog_subgoal1 *-> Prolog_subgoal2)
) :-
    !,
    compile_rholog_clause_body(
        Rholog_subgoal1, Subst_in, Subst, Prolog_subgoal1),
    compile_rholog_clause_body(
        Rholog_subgoal2, Subst, Subst_out, Prolog_subgoal2).
compile_rholog_clause_body(
    (Rholog_subgoal1; Rholog_subgoal2), Subst_in, Subst_out,
    (Prolog_subgoal1; Prolog_subgoal2)
) :-
    !,
    compile_rholog_clause_body(
        Rholog_subgoal1, Subst_in, Subst_out, Prolog_subgoal1),
    compile_rholog_clause_body(
        Rholog_subgoal2, Subst_in, Subst_out, Prolog_subgoal2).
compile_rholog_clause_body(!, Subst, Subst, !) :-
    !.
compile_rholog_clause_body(nl, Subst, Subst, nl) :-
    !.
compile_rholog_clause_body(write(X), Subst_Acc, Subst, Prolog_subgoal) :-
    !,
    term_to_applicative_form(
        write(X), Subst_Acc, _,  _, Subst, _, Prolog_style_subgoal_list),
    symbol_list_to_term(Prolog_style_subgoal_list, Prolog_subgoal).
compile_rholog_clause_body(
    Prolog_style_subgoal, Subst_in, Subst_out,
    (  symbol_list_to_term(Prolog_style_subgoal_list, Prolog_subgoal),
       Prolog_subgoal
    )
) :-
    !,
    term_to_applicative_form(
        Prolog_style_subgoal, Subst_in, _, [], Subst_out, _, 
        Prolog_style_subgoal_list).

%------------------------------------------------------------------------------
%   remove_true(+Pre_body, +Body, +Post_body, -Prolog_body)
%   Prholog light converts hedges into an internal representation. To represent
%   the nested structure of terms Prholog light uses the predicate append.
%   If there is no nested structure we do not want the unnecessary call true in
%   the converted code. So this predicate removes such calls. Furthermore if
%   no regular constraints need to be checked also the checkConstraints call 
%   will be removed. 
%------------------------------------------------------------------------------
remove_true(
    (Append1,Constraint1), Prolog_body, (Append2,Constraint2), Result
) :-
    remove_not_needed_calls([Append1,Constraint1] ,Output1),
    remove_not_needed_calls([Append2,Constraint2], Output2),
    append([Output1,[Prolog_body], Output2], Result_list),
    symbol_list_to_term(Result_list, Result),
    !.

%------------------------------------------------------------------------------
%   make_prolog_clause(
%       +Prholog_code, +Appends_constraints_in, -Appends_constraints_out, 
%       -Prolog_code)
%   Similar to remove_true but it works on Prholog clauses without a body.
%   TBD: Could be unified with remove true.
%------------------------------------------------------------------------------
make_prolog_clause(
    Clause_head, (Append1,Constraint1), (Append2,Constraint2), Clause_head
) :-
    remove_not_needed_calls([Append1,Constraint1,Append2,Constraint2],Output),
    Output = [],
    !.

make_prolog_clause(Clause_head, (Append1,Constraint1), (Append2,Constraint2), 
    (Clause_head :- Output_term)
) :-
    remove_not_needed_calls([Append1,Constraint1,Append2,Constraint2], Output),
    symbol_list_to_term(Output, Output_term),
    !.


%------------------------------------------------------------------------------
%   remove_not_needed_calls(+List_of_calls, -Filtered_list_of_calls)
%   Removes true and empty check_constraints calls
%------------------------------------------------------------------------------
remove_not_needed_calls([],[]).
remove_not_needed_calls([In_h|In_t], Output) :-
    (   In_h = true
    ;   In_h = check_constraints(_,Constraint_list),
        (    var(Constraint_list)
        ;    Constraint_list = []
        )
    ),
    remove_not_needed_calls(In_t, Output).
remove_not_needed_calls([In_h|In_t], [In_h|Output]) :-
    remove_not_needed_calls(In_t, Output).

%------------------------------------------------------------------------------
%   get_translated_literal(
%       +Type, +Str_name, +Str_args, +Lhs_list, +Rhs_list, +Constraint, -Trans)
%   If a strategy appears in the body this predicate converts it to 
%   Prolog code. Type equals either 'positive' or 'negative' depending which 
%   operator (==> or =\=> is used)
%------------------------------------------------------------------------------

get_translated_literal(
    positive, Str_name, [], Lhs_list, Rhs_list, Constraint, Trans
) :-
    var(Str_name),
    !,
    Trans = (
        get_strategy_name_arguments_from_the_applicative_form(
            Str_name, Str_name_new, Str_args_list_new),
        Subgoal =.. [
            Str_name_new,Str_args_list_new,Lhs_list,Rhs_list,Constraint
        ],
        Subgoal
    ).
get_translated_literal(
    positive, Str_name, Str_args, Lhs_list, Rhs_list, Constraint, Trans
):-
    !,
    Trans = (
        Subgoal =.. [Str_name,Str_args,Lhs_list,Rhs_list,Constraint],
        Subgoal
    ).
get_translated_literal(
    negative, Str_name, [], Lhs_list, Rhs_list, Constraint, Trans
):-
    var(Str_name),
    !,
    Trans = (
        get_strategy_name_arguments_from_the_applicative_form(
            Str_name, Str_name_new, Str_args_list_new
        ),
        Subgoal =.. [
            Str_name_new, Str_args_list_new, Lhs_list, Rhs_list, 
            Constraint
        ],
        \+ Subgoal
    ).
get_translated_literal(
    negative, Str_name, Str_args, Lhs_list, Rhs_list, Constraint, Trans
) :-
    !,
    Trans = (
        Subgoal =.. [Str_name,Str_args, Lhs_list,Rhs_list,Constraint],
        \+ Subgoal
    ).

%------------------------------------------------------------------------------
%   ?(+Query, -Result)
% 
%   ? is the top-level predicate that executes Rholog goals. 
%   Query is an Prholog query and Result is a list of Substitutions containing 
%   the solution of the given query.
%------------------------------------------------------------------------------

?(Query, Result) :-
    catch(
        (
            check_rholog_query_syntax(Query),
            compile_rholog_query(Query, Result)
        ),
        E,
        print_message(user_error,E)
    ).

compile_rholog_query(Strategy :: Lhs ==> Rhs, Result) :-
    !,
    compile_rholog_query(Strategy :: Lhs ==> Rhs where _, Result).
compile_rholog_query(Strategy :: Lhs ==> Rhs where Constraints, Result) :-
    !,
    compile_rholog_clause_body(
        Strategy :: Lhs ==> Rhs where Constraints, [], Subst_out, Compiled),
    !,
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).
compile_rholog_query(Strategy :: Lhs =\=> Rhs, Result) :-
    compile_rholog_query(Strategy :: Lhs =\=> Rhs where _, Result).
compile_rholog_query(Strategy :: Lhs =\=> Rhs where Constraints, Result) :-
    !,
    compile_rholog_clause_body(
        Strategy :: Lhs =\=> Rhs where Constraints, [], Subst_out, Compiled),
    !,
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).
compile_rholog_query((Query1,Query2), Result) :-
    !,
    compile_rholog_clause_body((Query1,Query2), [], Subst_out, Compiled),
    !,
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).
compile_rholog_query((Query1;Query2), Result) :-
    !,
    compile_rholog_clause_body((Query1;Query2), [], Subst_out, Compiled), 
    !,
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).
compile_rholog_query(Prolog_subquery, Result) :-
    !,
    compile_rholog_clause_body(Prolog_subquery, [], Subst_out, Compiled), 
    !,
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).

%------------------------------------------------------------------------------
%   hedge_to_term_lists_and_appends(
%       +Hedge_in, +Subst_in, +Appends_in, +Terms_in, -Subst_out, -Appends_out,
%       -Hedge_out)
%   hedge_to_term_lists_and_appends transforms the given hedge (Hedge_in) into
%   its inner representation Hedge_out. To do so Subst_in, which is a list 
%   of Substitutions is extended with one entry for each variable 
%   (individual, function and sequent) linking it with a Prolog variable.
%   This extended list is represented by Subst_out. Terms_in contains 
%   all already converted terms. Hedge_out is a list of all converted terms. 
%------------------------------------------------------------------------------

hedge_to_term_lists_and_appends(
    eps, Subst_acc, Appends_acc, Term_list_acc, Subst_acc, Appends_acc, 
    Term_list_acc
) :-
    !.
hedge_to_term_lists_and_appends(
    (Term_0,Term_1), Subst_acc, Appends_acc, Term_list_acc, Subst, Appends, 
    Term_list
) :-
    !,
    hedge_to_term_lists_and_appends(
        Term_1, Subst_acc, Appends_acc, Term_list_acc, Subst_acc_temp, 
        Appends_acc_temp, Term_list_acc_Temp
    ),
    hedge_to_term_lists_and_appends(
        Term_0, Subst_acc_temp, Appends_acc_temp, Term_list_acc_Temp,
        Subst, Appends, Term_list
    ).
hedge_to_term_lists_and_appends(
    Ind_var, Subst_acc, Appends_acc, Term_list_acc, Subst, Appends_acc, 
    [Prolog_var|Term_list_acc]
) :-
    atom(Ind_var),
    ind_var(Ind_var),
    !,
    prholog_var_to_prolog_var(Ind_var, Subst_acc, Prolog_var, Subst).
hedge_to_term_lists_and_appends(
    Seq_var, Subst_acc, true, Term_list_acc, Subst, true, Term_list
) :-
    atom(Term_list_acc),
    atom(Seq_var),
    seq_var(Seq_var),
    !,
    prholog_var_to_prolog_var(Seq_var, Subst_acc, Term_list, Subst).
hedge_to_term_lists_and_appends(
    Seq_var, Subst_acc, Appends_acc, Term_list_acc, Subst,
    my_append(Prolog_var,Term_list_acc,Term_list), Term_list
) :-
    atom(Appends_acc),
    atom(Seq_var),
    seq_var(Seq_var),
    !,
    prholog_var_to_prolog_var(Seq_var, Subst_acc, Prolog_var, Subst).

hedge_to_term_lists_and_appends(
    Seq_var, Subst_acc, Appends_acc, Term_list_acc, Subst,
    (my_append(Prolog_var,Term_list_acc,Term_list), Appends_acc),
    Term_list
) :-
    atom(Seq_var),
    seq_var(Seq_var),
    !,
    prholog_var_to_prolog_var(Seq_var, Subst_acc, Prolog_var, Subst).

hedge_to_term_lists_and_appends(
    Single_term, Subst_acc, Appends_acc, Term_list_acc, Subst, Appends,
    [Single_term_list|Term_list_acc]
) :-
    term_to_applicative_form(
        Single_term, Subst_acc, Appends_acc, [], Subst, Appends, 
        Single_term_list).


%------------------------------------------------------------------------------
%   my_append(?List1, ?List2, ?List3) 
%   It is a wrapper predicate needed for the modified trace
%------------------------------------------------------------------------------

my_append(A,B,C) :- append(A,B,C).

%------------------------------------------------------------------------------
%   term_to_applicative_form(
%       +Term, +Subst_in, +Appends_in, +Term_in, -Subst_out, -Appends_out,
%       -Hedge_out)
%   Similar to hedge_to_term_lists_and_appends but it works on terms not on
%   hedges.
%------------------------------------------------------------------------------

term_to_applicative_form(
    Ind_var_or_number, Subst_acc, Appends_acc, [], Subst, Appends_acc, 
    Single_term_list
) :-
    atomic(Ind_var_or_number),
    (   number(Ind_var_or_number)
    ;   ind_var(Ind_var_or_number)
    ;   ind_var_anonymous(Ind_var_or_number)
    ),
    !,
    transform_ind_var_or_number(
        Ind_var_or_number, Subst_acc, Single_term_list, Subst).
term_to_applicative_form(
    Single_term, Subst_acc, Appends_acc, Term_list_acc, Subst, Appends,
    apply_rholog_internal(Functor_translated, Arguments_list)
) :-
    Single_term =.. [Functor|Arguments],
    transform_fun_var_or_fun_symbol(
        Functor, Subst_acc, Functor_translated, Subst_acc_temp),
    pseudo_list_to_list(
        Arguments, Subst_acc_temp, Appends_acc, Term_list_acc, Subst, 
        Appends, Arguments_list).


%------------------------------------------------------------------------------
%   pseudo_list_to_list(
%       +Args_in, +Subst_in, +Appends_in, +Term_in, -Subst_out, 
%       -Appends_out, -Hedge_out)
%   Converts the arguments of a given term to its inner representation. 
%   Similar to term_to_applicative_form
%------------------------------------------------------------------------------

pseudo_list_to_list(
    [Seq_var], Subst_acc, Appends_acc, _, Subst, Appends_acc, Term_list
) :-
    atomic(Seq_var),
    seq_var(Seq_var),
    !,
    prholog_var_to_prolog_var(Seq_var, Subst_acc, Term_list, Subst).
pseudo_list_to_list(
    [Seq_var], Subst_acc, Appends_acc, _, Subst_acc, Appends_acc, _
) :-
    atomic(Seq_var),
    seq_var_anonymous(Seq_var),
    !.
pseudo_list_to_list(
    [Term|Other_Terms], Subst_acc, Appends_acc, Term_list_acc,
    Subst, Appends, Term_list
) :-
    pseudo_list_to_list(
        Other_Terms, Subst_acc, Appends_acc, Term_list_acc,
        Subst_acc_temp, Appends_acc_temp, Term_list_acc_Temp),
    hedge_to_term_lists_and_appends(
        Term, Subst_acc_temp, Appends_acc_temp,
        Term_list_acc_Temp, Subst, Appends, Term_list).
pseudo_list_to_list(
    [], Subst_acc, Appends_acc, Term_list_acc, Subst_acc, Appends_acc, 
    Term_list_acc).

%------------------------------------------------------------------------------
%   symbol_list_substitutions_to_term_substitutions(+Subst_in, -Subst_out)
%   Subst_in is a list of substitutions of the form Var_name ---> Var_value.
%   symbol_list_substitutions_to_term_substitutions converts each Var_value
%   from its inner representation back to a hedge.
%------------------------------------------------------------------------------

symbol_list_substitutions_to_term_substitutions([], []) :-
    !.
symbol_list_substitutions_to_term_substitutions(
    [Seq_var ---> T_l|Rest1], [Seq_var ---> T|Rest2]
) :-
    seq_var(Seq_var),
    !,
    symbol_list_to_term([T_l], T),
    symbol_list_substitutions_to_term_substitutions(Rest1, Rest2).
symbol_list_substitutions_to_term_substitutions(
    [X ---> T_l|Rest1], [X ---> T|Rest2]
) :-
    symbol_list_to_term(T_l, T),
    symbol_list_substitutions_to_term_substitutions(Rest1, Rest2).

list_of_symbol_lists_to_symbol_list([], []).
list_of_symbol_lists_to_symbol_list([X|T], [X|T1]):-
    var(X),
    !,
    list_of_symbol_lists_to_symbol_list(T, T1).
list_of_symbol_lists_to_symbol_list([apply_rholog_internal(X1,Y1)|T], [X|T1]) :-
    symbol_list_to_term(apply_rholog_internal(X1,Y1), X),
    !,
    list_of_symbol_lists_to_symbol_list(T, T1).
list_of_symbol_lists_to_symbol_list([X|T], [X|T1]):-
    !,
    list_of_symbol_lists_to_symbol_list(T, T1).

symbol_list_to_term(X, X):-
    var(X),
    !.
symbol_list_to_term([], eps):-
    !.
symbol_list_to_term([X1], X2):-
    !,
    symbol_list_to_term(X1, X2).
symbol_list_to_term([H1,H2|Tail], (X,Y)):-
    !,
    symbol_list_to_term(H1, X),
    symbol_list_to_term([H2|Tail], Y).
symbol_list_to_term(apply_rholog_internal(Functor,Args), Term):-
    !,
    list_of_symbol_lists_to_symbol_list(Args, Args1),
    Term =.. [Functor|Args1].
symbol_list_to_term(X, X).

%------------------------------------------------------------------------------
%   transform_ind_var_or_number(+Var_name, +Subst_in, ?Var_value, -Subst_out)
%   Adds the given Var_name ---> Var_value to Subst_in. Var_value may be 
%   initialized. Var_name is either a inidividual or a function variable.
%   Constants do not affect the substitution.
%------------------------------------------------------------------------------

transform_ind_var_or_number(Ind_var, Acc, X, Acc) :-
    member(Ind_var ---> X, Acc),
    !.
transform_ind_var_or_number(Ind_var, Acc, _, Acc) :-
    ind_var_anonymous(Ind_var),
    !.
transform_ind_var_or_number(Ind_var, Acc, X, [Ind_var ---> X|Acc]) :-
    ind_var(Ind_var),
    !.
transform_ind_var_or_number(Number, Acc, Number, Acc).

transform_fun_var_or_fun_symbol(Fun_var, Acc, X, Acc) :-
    member(Fun_var ---> X, Acc),
    !.
transform_fun_var_or_fun_symbol(Fun_var, Acc, _, Acc) :-
    fun_var_anonymous(Fun_var),
    !.
transform_fun_var_or_fun_symbol(Fun_var, Acc, X, [Fun_var ---> X|Acc]) :-
    fun_var(Fun_var),
    !.
transform_fun_var_or_fun_symbol(Fun_symbol, Acc, Fun_symbol, Acc).

%% see [Note 4]
prholog_var_to_prolog_var(Prholog_var,Subst_in,_,Subst_in) :-
    (   ind_var_anonymous(Prholog_var)
    ;   fun_var_anonymous(Prholog_var)
    ;   seq_var_anonymous(Prholog_var)
    ).
prholog_var_to_prolog_var(Prholog_var, Subst_in, Prolog_var, Subst_in):-
    member(Prholog_var ---> Prolog_var, Subst_in),
    !.
prholog_var_to_prolog_var(
    Prholog_var, Subst_in, Prolog_var, 
    [Prholog_var ---> Prolog_var|Subst_in]
).

%------------------------------------------------------------------------------
%   get_strategy_name_and_arguments(
%      +Strategy, +Subst_in, -Strategy_name, -Strategy_arguments_list,
%      -Subst_out, +Appends_in, -Appends_out)
%   Gets strategy name and arguments. Arguments are returned as term list.
%   Prolog lists should not be used in strategies.
%------------------------------------------------------------------------------

get_strategy_name_and_arguments(
    Strategy, Subst_in, Strategy_name, [], Subst_out, Appends_in, Appends_in
) :-
    atomic(Strategy),
    (   number(Strategy)
    ;   ind_var(Strategy)
    ;   ind_var_anonymous(Strategy)
    ),
    !,
    transform_ind_var_or_number(
        Strategy, Subst_in, Strategy_name, Subst_out).
get_strategy_name_and_arguments(
    Strategy, Subst_in, Strategy_name, [], Subst_out, Appends_in, Appends_in
) :-
    atomic(Strategy),
    !,
    transform_fun_var_or_fun_symbol(
        Strategy, Subst_in, Strategy_name, Subst_out).
get_strategy_name_and_arguments(
    Strategy, Subst_in, Strategy_name, Strategy_arguments_list, Subst_out,
    Appends_acc, Appends_out
) :-
    term_to_applicative_form(
        Strategy, Subst_in, Appends_acc, [], Subst_out, Appends_out, 
        apply_rholog_internal(Strategy_name, Strategy_arguments_list)).

%------------------------------------------------------------------------------
%   get_strategy_name_arguments_from_the_applicative_form(
%       +Applicative_form, -Strategy_name, -Strategy_arguments_list)
%   Get strategy name and arguments from the applicative form
%------------------------------------------------------------------------------

get_strategy_name_arguments_from_the_applicative_form(
    apply_rholog_internal(Strategy_name,Strategy_arguments_list),
    Strategy_name, Strategy_arguments_list
) :-
    !.
get_strategy_name_arguments_from_the_applicative_form(Strategy, Strategy, []).


%------------------------------------------------------------------------------
% reconstruct_strategy_name_and_arguments_from_applicative_form(+X, -Y). 
% TBD: remove if not used
%------------------------------------------------------------------------------
/*reconstruct_strategy_name_and_arguments_from_applicative_form([], []).
reconstruct_strategy_name_and_arguments_from_applicative_form(
    [apply_rholog_internal(F, Args)|Rest],[T|Rest_reconstructed]
):-
    !,
    reconstruct_strategy_name_and_arguments_from_applicative_form(
        Args, Args_reconstructed
    ),
    T=..[F| Args_reconstructed],
    reconstruct_strategy_name_and_arguments_from_applicative_form(
        Rest, Rest_reconstructed
    ).

reconstruct_strategy_name_and_arguments_from_applicative_form(
    [N|Rest],[N|Rest_reconstructed]
):-
    !,
    reconstruct_strategy_name_and_arguments_from_applicative_form(
        Rest, Rest_reconstructed
    ).
*/
%------------------------------------------------------------------------------
% ind_var(+Var_name), fun_var(+Var_name), seq_var(+Var_name) 
% Checks for individual, function and sequence variables.
%------------------------------------------------------------------------------

ind_var([]) :-
    !,
    fail.
ind_var([_|_]) :-
    !,
    fail.
ind_var(X) :-
    atom_concat(i_, _, X).

fun_var([]) :-
    !,
    fail.
fun_var([_|_]) :-
    !,
    fail.
fun_var(X):-
    atom_concat(f_, _, X).

seq_var([]) :-
    !,
    fail.
seq_var([_|_]) :-
    !,
    fail.
seq_var(X) :-
    atom_concat(s_, _, X).


%------------------------------------------------------------------------------
% ind_var_anonymous(+Var_name),
% fun_var_anonymous(+Var_name),
% seq_var_anonymous(+Var_name),
% Recognizing anonymous variables
%------------------------------------------------------------------------------

ind_var_anonymous(i_).
fun_var_anonymous(f_).
seq_var_anonymous(s_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary predicates for tracing by Klaus Reisenberger-Hagmayr
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% set_trace_properties
% sets global variables for the default strategies and initializes variables
% for strategies defined in the input file
set_trace_properties :- 
    nb_setval(lib_strategies,
        [all_answers,choice,closure,compose,first_all,first_one,
	 id,interactive,iterate,map, map_to_subhedges,nf,rewrite,
	 join, sel
	]),                        % see [Note 1]
    nb_setval(strategies, []),     % see [Note 2]
    nb_setval(neg_strategies, []). % see [Note 3] 

:-set_trace_properties.

% add_strategy(+Strategy)
% adds the strategy given by Strategy to the global variable strategies
add_strategy(Strategy) :-
    nb_getval(strategies, Old_list_of_strategies),
    not(member(Strategy, Old_list_of_strategies)),
    New_list_of_strategies = [Strategy|Old_list_of_strategies],
    nb_setval(strategies, New_list_of_strategies).
add_strategy(_).

% add_neg_strategy(+Strategy)
% adds the strategy given by Strategy to the global variable neg_strategies
add_neg_strategy(Strategy) :-
    nb_getval(neg_strategies, Old_list_of_strategies),
    not(member(Strategy, Old_list_of_strategies)),
    New_list_of_strategies = [Strategy|Old_list_of_strategies],
    nb_setval(neg_strategies, New_list_of_strategies).
add_neg_strategy(_).

