/******************************************************************************
  Name: constraints-l.pl
  Description: This file contains all predicates for handling regular
    expressions.
  Version: 1.0.0.0
  Author(s): Klaus Reisenberger-Hagmayer
  Created: 2015-11-17
  Last edited: 2015-11-17
  Last edited by: Klaus Reisenberger-Hagmayer
  Copyright:
  License:
  Notes:
    Note1: if the variable is not found (Out = false) the program gets stuck 
    Note2: if Out equals the empty set (or hole for contexts) we succeeded 
      otherwise the regex matches the input only up to a certain point
    Note3: is needed because seq_var crashes if Pattern is not an atom
******************************************************************************/
%------------------------------------------------------------------------------
%   hedge_to_arg_list(+Hedge, -Hedge_as_list)
%   Transforms a hedge to a list of terms
%------------------------------------------------------------------------------
hedge_to_arg_list((Hedge_h,Hedge_t), [Arg_list_h|Arg_list_t]) :-
    Arg_list_h = Hedge_h,
    hedge_to_arg_list((Hedge_t), Arg_list_t),
    !.
hedge_to_arg_list((Hedge), [Hedge]) :- !.
hedge_to_arg_list(Hedge, [Hedge]).

%------------------------------------------------------------------------------
%   check_constraints(+Substitution, +Constraint_list)
%   Searches the Substitution for the given constraint variables
%   and checks whether the regex is fullfilled or not. Constraint_list is 
%   a list of items of the form [Var, Regexp, Integer_flag]
%------------------------------------------------------------------------------
check_constraints(_, Constraint_list) :-
    var(Constraint_list),
    !.
check_constraints(_, []) :- !.
check_constraints(Subst, [H_constraint|T_constraint]) :-
    extract_constraints(H_constraint, Constraint_as_list),
    Constraint_as_list = [Var, Regexp, Integer_flag],
    find_subst(Subst, Var, Out),
    (   nonvar(Out) ->    %TBD: see [Note 1]
        (   break_point,    %DEBUG
            symbol_list_to_term([Out], Out_hedge), 
            % first the internal representation is transformed into a hedge
            hedge_to_arg_list(Out_hedge, Out_list_of_terms), 
            % this hedge is then transformed into a list of terms 
            % (it is easier to handle lists than hedges)
            check_single_sequent_regex(
                [Out_list_of_terms,Regexp,Integer_flag], Subst
            )
        )
    ;   true
    ),
    check_constraints(Subst, T_constraint),
    !.


%------------------------------------------------------------------------------
%   extract_constraints(+Subst_in, -Subst_out)
%   Auxilary predicate to convert the different types of constraints to 
%   the standard form [Var, Regexp, Integer_flag]
%------------------------------------------------------------------------------
extract_constraints(In, Out) :-
    In = [Var,Regexp],
    Out = [Var,Regexp,0],
    !.
extract_constraints(In, Out) :-
    [In] = [Var in Regexp],
    Out = [Var,Regexp,0],
    !.
extract_constraints(In, In).


%------------------------------------------------------------------------------
%   break_point
%   Auxilary predicate which can be inserted into the code and used as a 
%   spypoint to make debugging easier
%------------------------------------------------------------------------------
break_point :- true.


%------------------------------------------------------------------------------
%   find_subst(+Substitution, +Var_name, -Var_value)
%   Checks Substitution if it contains an entry (Var_name ---> Var_value)  
%   with name Var_name and returns its value
%------------------------------------------------------------------------------
find_subst([], _, false).
find_subst([H_subst|T_subst], Var_name, Out) :-
    (   H_subst = Var_name ---> Value ->
        Out = Value
    ;   find_subst(T_subst, Var_name, Out)
    ),
    !.


%   Checks regexp for a single sequent variable
%------------------------------------------------------------------------------
%   check_single_sequent_regex(+Constraint, +Subst)
%   Calls check_regex for Constraint. Constraint is not a regular constraint 
%   over contexts.
%------------------------------------------------------------------------------
check_single_sequent_regex([Var,_,Integer_flag], _) :-
    Integer_flag = 0,
    Var = [],
    !.
check_single_sequent_regex([Var,Regexp,_], Subst) :-
    check_regex(Var, Regexp, false, Subst),
    !.


%   Checks a regular expression for a single variable
%------------------------------------------------------------------------------
%   check_regex(+Var_name, +Regexp, +Is_context, +Subst)
%   check_regex tests whether or not Regexp is fullfilled for the item of Subst
%   with name Var_name
%------------------------------------------------------------------------------
check_regex(In, Pattern, Is_ctx, Subst) :-
    check_regex_item(In, Pattern, Out, Is_ctx, Subst),   %% see [Note 6]
    (   Is_ctx ->
        Out = hole
    ;   Out = []
    ).    %% see [Note 2]


%------------------------------------------------------------------------------
%   check_regex_item(+Var_name, +Regexp, -Out, +Is_context, +Subst)
%   Checks whether a given hedge given by Var_name and Subst term by term 
%   satisfies Regex. With each application of check_regex_item the input hedge
%   is reduced and returned as Out. 
%------------------------------------------------------------------------------

%   Star: The inner pattern must match the beginning of the hedge given by 
%   Var_name 0 or more times 
%   (for this we use the auxilary predicate check_regex_multiple_times)
check_regex_item(In, Pattern, Out, Is_ctx, Subst) :-
    Pattern = sstar(Inner_pattern),
    !,
    check_regex_multiple_times(In, Inner_pattern, Out, Is_ctx, Subst).

% Or: One of the inner patterns must match the beginning of input
check_regex_item(In, Pattern, Out, Is_ctx, Subst) :-
    Pattern = sor(Inner_p1, Inner_p2),
    (   check_regex_item(In, Inner_p1, Out, Is_ctx, Subst)
    ;   check_regex_item(In, Inner_p2, Out, Is_ctx, Subst)
    ).

% Concat: All items of concat must match the beginning of input
check_regex_item(In, sconc([]), In, _, _).
check_regex_item(In, sconc(eps), In, _, _).
check_regex_item(In, Pattern , Out, Is_ctx, Subst) :-
    Pattern = sconc(T1),
    check_regex_item(In, T1, Out, Is_ctx, Subst).
check_regex_item(In, Pattern , Out, Is_ctx, Subst) :-
    Pattern = sconc(T1, T2),
    check_regex_item(In, T1, Out_aux, Is_ctx, Subst),
    check_regex_item(Out_aux, sconc(T2), Out, Is_ctx,Subst).
check_regex_item(In, Pattern, Out, Is_ctx, Subst) :-
    Pattern = (_, _),
    convert_hedge_to_pattern(Pattern, Sconc_pattern),
    check_regex_item(In, Sconc_pattern, Out, Is_ctx, Subst).

% Here we check the single non anonymous sequent vars (order matters!)
check_regex_item(In, Pattern, Out, Is_ctx, Subst) :-
    atom(Pattern), % see [Note 3]
    seq_var(Pattern),
    Pattern \= s_,
    find_subst(Subst, Pattern, Out_aux),
    nonvar(Out_aux),
    symbol_list_to_term([Out_aux], Out_hedge),
    convert_hedge_to_pattern(Out_hedge, Seq_pattern),
    check_regex_item(In, Seq_pattern, Out, Is_ctx, Subst),
    !.

% If Pattern is an anonymous sequent variable an arbitrary number
% of terms at the beginning of In are removed
check_regex_item(In, Pattern, Out, _,_) :-
    seq_var_anonymous(Pattern),
    append(_, Out, In).

% If Pattern is an anonymous individual variable 
% one term at the beginning of In is removed
check_regex_item(In, Pattern, Out, _,_) :-
    ind_var_anonymous(Pattern),
    append([_], Out, In),
    !.

% Pattern is an individual variable
check_regex_item(In, Pattern, Out, _, Subst) :-
    atom(Pattern),
    ind_var(Pattern),
    find_subst(Subst, Pattern, Out_ind_var),
    append([Out_ind_var], Out, In),
    !.

% Pattern is a ground term or constant
check_regex_item(In, Pattern, Out, _, _) :-
    append([Pattern], Out, In).

% The following rule is for checking regular constraints
% over contexts. Since contexts are not used in 
% Prholog-Light they are not needed here !!!works only 
% with a different internal representation
/*
check_regex_item(In, Pattern, Out, Is_ctx,Subst) :-
    (   Is_ctx -> 
        (
            groundTermToInner_representation(In, Internal_rep_in),
            groundTermToInner_representation(Pattern, Internal_rep_pattern),
            append(
                [Ctx1,[arg_beg(Level),hole,arg_end(Level)],Ctx2], 
                Internal_rep_pattern
            ),
            append([Ctx1,Inner_context,Ctx2], Internal_rep_in),
            Inner_context = [arg_beg(Level)|_],
            relevel(Inner_context,-Level,[],Out_aux),
            transform_subst_back_aux(test,Out_aux,Out)
        )
    ;   (   Pattern = s_ ->
            append(List_dummy,Out,In)
        ;   (   Pattern = i_ ->
                append([Dummy_term],Out,In)
            ;   append([Pattern],Out,In)
            )
        )
    ).
*/

%------------------------------------------------------------------------------
%   convert_hedge_to_pattern(+Hedge,-Pattern)
%   Auxilary predicate to convert a sequence into a term which is 
%   recognized by check_regex
%------------------------------------------------------------------------------
convert_hedge_to_pattern(eps, sconc([])) :- !.
convert_hedge_to_pattern((Part1,Part2), sconc(Part1,Converted_part2)) :-
    convert_hedge_to_pattern(Part2, Converted_part2),
    !.
convert_hedge_to_pattern((Part), Part).    


%------------------------------------------------------------------------------
%   check_regex_multiple_times(+Var_name, +Regexp, -Out, +Is_context, +Subst)
%   Auxilary predicate for star patterns. The given pattern can occur 0 or more
%   times
%------------------------------------------------------------------------------
check_regex_multiple_times(In, Pattern, Out, Type, Subst) :-
    check_regex_item(In, Pattern, Out_aux, Type, Subst),
    check_regex_multiple_times(Out_aux, Pattern, Out, Type, Subst),
    !.
check_regex_multiple_times(In, _, In, _, _). 

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Aux predicates for testing regular constraints (they are only needed 
% for regex over contexts which will not be used here)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Has a standard term as input and gives as output its inner representation 
% (a flat list which represents the term)
groundTermToInner_representation(Input, Output) :-
    groundTermToInner_representationAux([Input], Output, 0).


%[In_h|In_t]: A list of standard terms which is transformed to its inner 
% representation
%Level: The current level of operation
%[Out_h|Out_t]: The output

groundTermToInner_representationAux([], [], _) :- !.
groundTermToInner_representationAux([In_h|In_t], Out, Level) :-
    In_h =.. [Functor|Args],
    Next_level is Level + 1,
    groundTermToInner_representationAux(Args, Transformed_args, Next_level),
    (
        Args = [] -> 
            append(
                [[arg_beg(Level), Functor ,arg_end(Level)]],Out_h_transformed);
            append(
                [[arg_beg(Level), Functor, br_beg(Next_level)], 
                Transformed_args, 
                [br_end(Next_level) ,arg_end(Level)]],
                Out_h_transformed)
    ),
    groundTermToInner_representationAux(In_t,Out_t_transformed,Level),
    append([Out_h_transformed,Out_t_transformed],Out). 

relevel([],_,Acc,Out) :-
    reverse(Acc,Out).
relevel([In_h|In_t],Level,Acc,Out) :-
    In_h =.. List_rep,
    List_rep = [Functor, Old_level],
    New_level is Old_level + Level,
    Out_h =.. [Functor, New_level],
    Out_aux = [Out_h|Acc],
    relevel(In_t,Level,Out_aux,Out),!.
relevel([In_h|In_t],Level,Acc,Out) :- 
    relevel(In_t,Level,[In_h|Acc],Out),!.

%------------------------------------------------------------------------------
% transform_back(Subst_in, Subst_out) TBD: Check if still necessary
%
% Transforms Subst_in (which is still in the internal representation) back to 
% the usual representation
%
% i.e.:
% transform_subst_back_aux(_G1743,[arg_beg(0),is,br_beg(1),arg_beg(1),_G1313,
% arg_end(1),arg_beg(1),-,br_beg(2),arg_beg(2),3,arg_end(2),arg_beg(2),1,
% arg_end(2),br_end(2),arg_end(1),br_end(1),arg_end(0)],_G1745).
%------------------------------------------------------------------------------

transform_subst_back([],[]).
transform_subst_back([In_h|In_t],[Out_h|Out_t]) :-
    Var_name ---> Inner_rep = In_h,
    %% relevel_inner_rep(Inner_rep,0,[],Inner_rep_releveled),
    transform_subst_back_aux(Var_name,Inner_rep,Usual_rep),
    Out_h = Var_name ---> Usual_rep,
    transform_subst_back(In_t, Out_t),!.

transform_subst_back_aux(_,[],eps) :- !.
transform_subst_back_aux(Var_name,In,Out) :-
    seq_var(Var_name),
    transform_subst_back_aux_args(In,Out_list),
    symbol_list_to_term(Out_list,Out),!.

transform_subst_back_aux(_,In,In) :-
    atom(In);
    number(In).
transform_subst_back_aux(_,[In],In) :- !.
transform_subst_back_aux(_,In,Out) :-
    append([[Functor, br_beg(Level)],Inner_part,[br_end(Level)]],In),!,
    transform_subst_back_aux_args(Inner_part,Args),
    append([Functor],Args,Term_list),
    Out =.. Term_list.
transform_subst_back_aux(_,In,Out) :-
    append([[arg_beg(Level)],Inner_part,[arg_end(Level)]],In),!,
    transform_subst_back_aux(t,Inner_part,Out),!.

transform_subst_back_aux_args([],[]).
transform_subst_back_aux_args(
    [arg_beg(Level),Inner_part,arg_end(Level)|Tail],[Inner_part|Out_t]
) :-
    transform_subst_back_aux_args(Tail,Out_t).
transform_subst_back_aux_args(In,[Out_h|Out_t]) :-
    append([[arg_beg(Level)],Inner_part,[arg_end(Level)],Tail],In),!,
    transform_subst_back_aux(t,Inner_part,Out_h),
    transform_subst_back_aux_args(Tail,Out_t),!.

*/
