/******************************************************************************
  Name: library-l.pl
  Description: Defines the default strategies included in Prholog-light.
  Version: 1.0.0.1
  Author(s): Besik Dundua
  Created: 2009-09-09
  Last edited: 2016-05-08
  Last edited by: Temur Kutsia
  Copyright:
  License:
  Notes:
    - Constraints is not used in this file but it is necessary for the 
      modified trace.
******************************************************************************/

%--------------------------------------------------------------------------
%   all_answers(+Strategy, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   all_answers returns a hedge consisting of all answers computed by 
%   Strategy for Input_hedge. To distinguish from each other, each answer 
%   is wrapped in the 'ans' constructor. Output_hedge must then match 
%   the result hedge. Can be seen as an analog of findall for PrhoLog. 
%   Exactly one strategy must be provided.
%--------------------------------------------------------------------------

all_answers([Strategy], Lhs, Rhs, Constraint) :-
    get_strategy_name_arguments_from_the_applicative_form(Strategy, Str_name, 
    Str_args),
    Strategy_appication =.. [Str_name,Str_args,Lhs,Temp,Constraint],
    findall(
        apply_rholog_internal(ans,Temp),
        Strategy_appication,
        Rhs).


%------------------------------------------------------------------------------
%   choice(+Strategies, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   Chooses a strategy from Strategies that successfully applies to
%   Input_hedge under Constraints. Output_hedge matches the result. 
%   With backtracking one can get all possible results. 
%   If there is no strategy in Strategies that succeeds, choice fails.
%------------------------------------------------------------------------------

choice([Strategy1|_], Lhs, Rhs, Constraints) :-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy1, Str1_name, Str1_args),
    Strategy_appication =.. [Str1_name,Str1_args,Lhs,Rhs,Constraints],
    Strategy_appication.
choice([_|Strategies], Lhs, Rhs, Constraints) :-
    choice(Strategies, Lhs, Rhs, Constraints).


%------------------------------------------------------------------------------
%   closure(+Strategy, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   Applies Strategy to Input_hedge under Constraints repeteadly as long as 
%   possible, one after the other. Each time, Output_hedge matches the result. 
%   Results of applications are returned one by one, as they are computed via 
%   backtracking. Exactly one strategy must be provided.
%------------------------------------------------------------------------------

closure([_], X, X, _).
closure([Strategy], Lhs, Rhs, Constraints):-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy, Str_name, Str_args),
    Strategy_appication =.. [Str_name,Str_args,Lhs,Rhs_temp,Constraints],
    Strategy_appication,
    closure([Strategy], Rhs_temp, Rhs, Constraints).


%------------------------------------------------------------------------------
%   compose(+Strategies, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   compose first applies the first strategy from Strategies to the Input_hedge
%   under Constraints. To its result, the second strategy is applied and so on.
%   Output_hedge matches the final result. It fails if one of the strategies 
%   fails. All possible results can be obtained by backtracking. 
%   At least two strategies must be provided.
%------------------------------------------------------------------------------

compose([Strategy1, Strategy2], Lhs, Rhs, Constraints):-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy1, Str1_name, Str1_args),
    Strategy_appication1=..[Str1_name,Str1_args,Lhs,Rhs_temp,Constraints],
    Strategy_appication1,
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy2, Str2_name, Str2_args),
    Strategy_appication2=..[Str2_name,Str2_args,Rhs_temp,Rhs,Constraints],
    Strategy_appication2.
compose([Strategy1, Strategy2, Strategy3|Strategies], Lhs, Rhs, Constraints) :-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy1, Str1_name, Str1_args),
    Strategy_appication1=..[Str1_name,Str1_args,Lhs,Rhs_temp,Constraints],
    Strategy_appication1,
    compose([Strategy2, Strategy3|Strategies], Rhs_temp, Rhs, Constraints).


%------------------------------------------------------------------------------
%   first_all(+Strategies, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   first_all tries to apply the first strategy of Strategies to Input_hedge 
%   (under Constraints). If this fails, it tries the next one and so on. 
%   When a strategy is found that succeeds to Input_hedge, first_all returns
%   all answers computed by it, via backtracking. If no strategy succeeds, 
%   first_all fails.
%------------------------------------------------------------------------------

first_all([Strategy1|Strategies], Lhs, Rhs, Constraints) :-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy1, Str1_name, Str1_args),
    Strategy_appication =.. [Str1_name,Str1_args,Lhs,Rhs, Constraints],
    (   Strategy_appication *->
        true
    ;   first_all(Strategies, Lhs, Rhs, Constraints)
    ).


%------------------------------------------------------------------------------
%   first_one(+Strategies, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   first_one behaves similarly to first_all, with the difference that when
%   a strategy is found that succeeds to Input_hedge, first_one returns only
%   one answer computed by it.
%------------------------------------------------------------------------------

first_one([Strategy1|_], Lhs, Rhs, Constraints):-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy1, Str1_name, Str1_args),
    Strategy_appication =.. [Str1_name,Str1_args,Lhs,Rhs,Constraints],
    Strategy_appication,
    !.
first_one([_|Strategies], Lhs, Rhs, Constraints):-
    first_one(Strategies, Lhs, Rhs, Constraints).
   
     
%------------------------------------------------------------------------------
%   id(+Strategies, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   id succeeds if Output_hedge matches Input_hedge. Otherwise it fails. 
%   The list of Strategies is empty.
%------------------------------------------------------------------------------

id([], X, X, _).



%------------------------------------------------------------------------------
%   interactive([], +Input_hedge, ?Output_hedge, ?Constraints)
%
%   It starts interaction with the user, asking her to provide a strategy, 
%   which is then applied to the Input_hedge. The process is repeated further 
%   so that the output hedge of the previous application becomes the input for 
%   the new strategy privided by the user, and so on. The interactive process 
%   stops when the user types 'finish'. Output_hedge matches the input hedge 
%   that was there at the moment when 'finish' was typed. It fails when the 
%   user-provided strategy fails for the current input hedge.
%------------------------------------------------------------------------------

interactive([], Lhs, Rhs, Constraints):-
    symbol_list_to_term(Lhs, Lhs_hedge),
    writef('The hedge to be transformed\n\t%w.\n',[Lhs_hedge]),
    write('Type the strategy name (or \'finish\' to stop transformation): '),
    read(Strategy),
    nl,
    Strategy \= finish,
    !,
    get_strategy_name_and_arguments(Strategy, [], Str_name, Str_args_list, _),
    Strategy_appication =.. [Str_name, Str_args_list, Lhs, Temp, Constraints],
    Strategy_appication,
    !,
    interactive([], Temp, Rhs, Constraints).

interactive([], X, X, _).

        
%------------------------------------------------------------------------------
%   iterate([+Strategy, +N], +Input_hedge, ?Output_hedge, ?Constraints)
%
%   Applies Strategy repeatedly to Input_hedge N times under Constraints. 
%   Output_hedge matches the result. Fails if one of the applications fails.
%------------------------------------------------------------------------------


iterate([Strategy, 1], Lhs, Rhs, Constraints):-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy, Str_name, Str_args),
    Strategy_appication=..[Str_name,Str_args,Lhs,Rhs, Constraints],
    Strategy_appication.
iterate([Strategy, N], Lhs, Rhs, Constraints):-
    N>1,
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy, Str_name, Str_args),
    Strategy_appication=..[Str_name,Str_args,Lhs,Rhs_temp, Constraints],
    Strategy_appication,
    N1 is N-1,
    iterate([Strategy, N1], Rhs_temp, Rhs, Constraints).


%------------------------------------------------------------------------------
%   join([], +Input_hedge, ?Output_hedge, ?Constraints)
%
%   Input_hedge is a hedge consisting of two terms. When the terms have the
%   same top function symbol, Output_hedge matches the term with the same
%   top function symbol, whose areguments are the arguments of the first term
%   followed by the arguments of the second term. Fails when the input terms
%   do not have the same head.
%------------------------------------------------------------------------------

join([], [apply_rholog_internal(H,L1), apply_rholog_internal(H,L2)],
    [apply_rholog_internal(H,L)], _
) :-
    my_append(L1, L2, L).




%------------------------------------------------------------------------------
%   map(+Strategy, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   Maps Strategy to each term of Input_hedge (under Constraints). 
%   Output of Strategy for a input term can be a hedge (or, in particular, 
%   a single term). Output_hedge matches the hedge composed from 
%   Strategy outputs for each term in Input_hedge. It fails as soon as the 
%   application of Strategy to a term from Input_hedge fails.
%
%   When Input_hedge is empty, Output_hedge should be empty as well.
%------------------------------------------------------------------------------

map([Strategy], Lhs, Rhs, Constraints):-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy, Str_name, Str_args),
    map(Str_name, Str_args, Lhs, [], Rhs, Constraints).
map(_, _, [], Acc, Acc, _).
map(Str_name, Str_args, [First|Rest], Acc, Rhs, Constraints):-
    Strategy_appication =..
        [Str_name,Str_args,[First],First_result,Constraints],
    Strategy_appication,
    my_append(Acc, First_result, New_acc),
    map(Str_name, Str_args, Rest, New_acc, Rhs, Constraints).


%------------------------------------------------------------------------------
%   map_to_subhedges(+Strategy, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   Similar to map above, with the only difference that Strategy is mapped to 
%   not necesarily single terms from Input_hedge, but to nonempty subhedges of 
%   it.
%
%   When Input_hedge is empty, Output_hedge should be empty as well.
%------------------------------------------------------------------------------

map_to_subhedges(Str, Lhs, Rhs, Constraints):-
    map_to_subhedges(Str, Lhs, [], Rhs, Constraints).
map_to_subhedges(_, [], Acc, Acc, _).
map_to_subhedges([Strategy], [L|Lhs], Acc, Rhs, Constraints) :-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy, Str_name, Str_args),
    append([L|List1], List2, [L|Lhs]),
    Strategy_appication =..
        [Str_name,Str_args,[L|List1],LList1_Results,Constraints],
    Strategy_appication,
    append(Acc, LList1_Results, New_acc),
    map_to_subhedges([Strategy], List2, New_acc, Rhs, Constraints).


%------------------------------------------------------------------------------
%   nf(+Strategy, +Input_hedge, ?Output_hedge, ?Constraints)
%
%   Computes the normal form of application of Strategy to Input_hedge under
%   Constraints. Normal form is computed by applying Strategy to Input_hedge
%   repeatedly as long as the application succeeds. Whevener it fails, the 
%   current input hedge is returned and Output_hedge matches it. Exactly one 
%   strategy must be provided.
%------------------------------------------------------------------------------

nf([Strategy], Lhs, Rhs, Constraint):-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy, Str_name, Str_args),
    nf_rholog_internal(Str_name, Str_args, Lhs, Rhs, Constraint).
nf_rholog_internal(Str_name, Str_args, Lhs, Rhs, Constraint):-
    Strategy_appication =.. [Str_name,Str_args,Lhs,Temp,Constraint],
    Strategy_appication,
    nf_rholog_internal(Str_name, Str_args, Temp, Rhs, Constraint).
nf_rholog_internal(Str_name, Str_args, H, H,Constraint):-
    Strategy_appication =.. [Str_name,Str_args,H,_,Constraint],
    \+(Strategy_appication).


%------------------------------------------------------------------------------
%   rewrite(+Strategy, +Input_term, ?Output_term, ?Constraints)
%
%   Rewrites Input_term by Strategy. Output_term matches the result. Rewriting
%   means that the a subterm of Input_term (not necessarily the entire 
%   Input_term) is transformed by Strategy. It fails if Strategy fails for all
%   subterms of Input_term. All answers can be computed by backtracking.
%   Exactly one strategy must be provided. Input and output are single terms,
%   not arbitrary hedges.
%------------------------------------------------------------------------------

rewrite([Strategy], Term, Term_rewritten, Constraints) :-
    get_strategy_name_arguments_from_the_applicative_form(
        Strategy, Str_name, Str_args),
    Strategy_appication =..[
        Str_name, Str_args, Term, Term_rewritten, Constraints],
    Strategy_appication.
rewrite(
    [Strategy], [apply_rholog_internal(Functor,Args)],
    [apply_rholog_internal(Functor,Args_rewritten)], Constraints
) :-
    rewrite_list([Strategy], Args, Args_rewritten, Constraints).

rewrite_list(
    [Strategy], [Arg1|Other_args], [Arg1_rewritten|Other_args], Constraints
) :-
    rewrite([Strategy],[Arg1],[Arg1_rewritten], Constraints).
rewrite_list(
    [Strategy], [Arg1|Other_args], [Arg1|Other_args_rewritten], Constraints
) :-
    rewrite_list([Strategy], Other_args, Other_args_rewritten, Constraints).


%------------------------------------------------------------------------------
%   sel([], +Input_hedge, ?Output_hedge, ?Constraints)
%
%   Splits the input hedge into one element of the hedge and the remaining 
%   elements. This strategy is usually used with a predicate which restricts 
%   the single element in some way. By doing so we can extract a term with a 
%   certain property from the hedge. 
%------------------------------------------------------------------------------

sel([],  L, [X|L1], _) :-
    my_select(X, L, L1).

% my_select(?Elem, ?List, ?Rest)
% A wrapper predicate for select to be able to exclude it from the modified
% trace
my_select(A, B, C) :- select(A, B, C).
